/* $Id: socket.c,v 1.3 2000/01/15 18:00:49 alinden Exp $ */

/*
*  File             : socket.c
*  Written by       : alinden@gmx.de
*  Portability      : Should be POSIX compliant
*  Copyright:       : GPL
*
*  DESCRIPTION: Hmm... looks like socket related stuff
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

void log_info(char *);
void log_warn(char *);
void update_ctab(char *);
char *cpstr(size_t, char *, ...);
void restore_sigs(void);
void msgout(char *);
void storequeue(CONFIG *, char);
void run_proc(CONFIG *);
void run_procs(CONFIG *);
void flushjob(CONFIG *);
time_t runtab(CONFIG *, time_t);
int cmd_queued(CONFIG *); 

void sig_nop(int signo) { return; }

void outfifo(CMD *inb)   /* send joblist to socket */
{
	pid_t pid;
	char *username;
	CONFIG *tbuf;
	FILE *ffifo=NULL;
	char cmain;
	char *tmstr = "%b %d %a %H:%M";
	time_t ltm,lth=0;
	char tmbuf[20];
	int jobs,fjobs,lstall,rall,jrun,jqueue,sjobs,i;

	if(inb->ctab == NULL || inb->par == NULL)
	{
		log_warn("outfifo(): missing parameter");
		return;
	}	
	
	if((pid = fork()) < 0)
	{
		log_warn("can't fork");
		return;
	}

	if(pid)
		return;

	restore_sigs();

	lstall = rall = jrun = jqueue = 0;

	if(strcmp(inb->ctab+2, NCTROOT) == 0)
	{
		username = "root";
		cmain = 1;
	}
	else
	{
		username = inb->ctab+2;
		cmain = 0;
	}	

	if(cmain)
	{
		if(strcmp(inb->par, "*all*") == 0)
		{
			lstall = 1;
		}	
		else if(strcmp(inb->par, "*rall*") == 0)
		{
			lstall = rall = 1;
		}	
	}

	if(! lstall)
	{
		if(strcmp(inb->par, "*run*") == 0)
		{
			jrun = 1;
		}
		else if(strcmp(inb->par, "*queue*") == 0)
		{
			jqueue = 1;
		}
	}		

	if((ffifo = fdopen(svsock, "w")) == NULL)
		exit(0);

	tbuf = cblock;

	if(lstall == 0)
	{
		while(tbuf != NULL)  /* find user */
		{
			if((cmain && tbuf->cmain) || 
			  (! cmain && ! tbuf->cmain && strcmp(username, tbuf->username) == 0))
				break;

			tbuf = tbuf->nuser;
		}
	}
	
	if(tbuf == NULL)
	{
		if(lstall == 0) 
			fprintf(ffifo, "No jobs for user %s\n", username);
		else
			fprintf(ffifo, "No jobs\n");

		fclose(ffifo);
		exit(0);
	}

	if(lstall == 0 && ! jrun && ! jqueue)
		fprintf(ffifo, "Jobs for user %s\n", username);

	time(&ltm);
		
	jobs=fjobs=sjobs=0;
	while(1)
	{
		if(lstall && tbuf->stblock)
		{
			sjobs++;
			jobs = 0;
		}	
		jobs++;

		if(lstall == 0)
		{
			if((jrun == 0 && jqueue == 0 && ! (tbuf->jobid != NULL && fnmatch(inb->par, tbuf->jobid, 0) == 0) && 
			fnmatch(inb->par, tbuf->cmdstr, 0)) || (jrun && jqueue == 0 && ! tbuf->pid) || (jrun == 0 && jqueue && 
			((! cmd_queued(tbuf) && ! tbuf->swpdelay && ! tbuf->qdelay) || tbuf->pid)))
			{
				tbuf = tbuf->nblock;

				if(tbuf == NULL || tbuf->stblock)
					break;

				continue;	
			}
		}
		else if(rall && ! tbuf->pid)
		{
			tbuf = tbuf->nblock;
	
			if(tbuf == NULL)
				break;

			continue;	
		}
		
		if((lstall == 0 && ! jrun && ! jqueue) || fjobs)
			fprintf(ffifo, "-----------------------------------------\n");

		fjobs = 1;	
		
		fprintf(ffifo, "Number : ");

		if(lstall)
			fprintf(ffifo, "%d/", sjobs);

		fprintf(ffifo, "%d\n", jobs);
		fprintf(ffifo, "%s\n", cpstr(79, "Command: %s", tbuf->cmdstr));
		
		if(cmain)
		{
			fprintf(ffifo, "User   : %s", tbuf->username);
			
			if(lstall == 1 && tbuf->cmain && strcmp(tbuf->username, "root"))
				fputs(" (from root)\n", ffifo);
			else
				fputc('\n', ffifo);
		}
		
		fprintf(ffifo, "Subject: %s\n", (tbuf->jobid)?(tbuf->jobid):("<default>"));
		fprintf(ffifo, "Options: ");

		i = 0;

		if(tbuf->period)
		{
			fputc('U', ffifo);
			i = 1;
		}	
		if(tbuf->dopt)
		{
			fputc('D', ffifo);
			i = 1;
		}

		if(tbuf->bopt)
		{
			fputc('B', ffifo);	
			i = 1;
		}

		if(tbuf->lopt)
		{
			fputc('L', ffifo);	
			i = 1;
		}

		if(tbuf->ropt)
		{
			fputc('R', ffifo);
			i = 1;
		}

		if(tbuf->qopt)
		{
			fputc('Q', ffifo);
			i = 1;
		}

		if(tbuf->atmout)
		{
			fputc('A', ffifo);
			i = 1;
		}

		if(tbuf->htmout)
		{
			fputc('H', ffifo);
			i = 1;
		}

		if(tbuf->mopt)
		{
			fputc('M', ffifo);
			i = 1;
		}

		if(i == 0)
			fprintf(ffifo, "<none>");

		fputc('\n', ffifo);
		
		fprintf(ffifo, "Status : ");

		if(tbuf->swpdelay)
		{
			fprintf(ffifo, "DELAYED\n");
		}
		else if(tbuf->pid)
		{
			fprintf(ffifo, "RUN/PID: %d\n", tbuf->pid);

		} 
		else if(cmd_queued(tbuf))
		{
			fprintf(ffifo, "QUEUED\n");
			
		} else
		{
			fprintf(ffifo, "WAIT");
			
			if(tbuf->qdelay)
			{
				fprintf(ffifo, "/Q\n");
			}
			else if(tbuf->lastexec)
			{
				if(tbuf->exite)
					fprintf(ffifo, "/EXIT ERROR\n");
				else
					fprintf(ffifo, "/EXIT %d\n", tbuf->exitc);
			}
			else fputc('\n', ffifo);
		}

		fprintf(ffifo, "Last   : ");

		if(tbuf->lastexec == 0)
			fprintf(ffifo, "<never>\n");
		else
		{
			strftime(tmbuf, 20, tmstr, localtime(&tbuf->lastexec));
			lth = (ltm-tbuf->lastexec)/(60*60);
			fprintf(ffifo, "%s  +(%dh %.0fm)\n", tmbuf, (int)lth, (((float)ltm-(float)tbuf->lastexec)/60)-(float)lth*60);
		}

		fprintf(ffifo, "Next   : ");

		if(tbuf->nextexec == -1)
			fprintf(ffifo, "<?>\n");
		else
		{
			strftime(tmbuf, 20, tmstr, localtime(&tbuf->nextexec));
			lth = (tbuf->nextexec-ltm)/(60*60);
			fprintf(ffifo, "%s  -(%dh %.0fm)\n", tmbuf, (int)lth, (((float)tbuf->nextexec-(float)ltm)/60)-(float)lth*60);
		}
		
		tbuf = tbuf->nblock;

		if(tbuf == NULL || (! lstall && tbuf->stblock))
			break;

	}

	if(lstall == 0 && ! jrun && ! jqueue && fjobs == 0)
		fprintf(ffifo, "No match with %s\n", inb->par);

	fclose(ffifo);

	exit(0);
}

CONFIG *findblk(CMD *inb, char **username)   /* find entry */
{
	CONFIG *tbuf;
	char cmain,*str=NULL;
	int job,i;
	
	if(inb->ctab == NULL || inb->par == NULL)
	{
		log_warn("findblk(): missing parameter");
		return(NULL);
	}
	
	if(strcmp(inb->ctab+2, NCTROOT) == 0)
	{
		*username = "root";
		cmain = 1;
	}
	else
	{
		*username = inb->ctab+2;
		cmain = 0;
	}	

	tbuf = cblock;

	while(tbuf != NULL)  /* find user */
	{
		if((cmain && tbuf->cmain) || 
		  (! cmain && ! tbuf->cmain && strcmp(*username, tbuf->username) == 0))
			break;

		tbuf = tbuf->nuser;
	}

	if(tbuf == NULL)
	{
		str = cpstr(256, "No jobs for user %s\n", *username);
		write(svsock, str, strlen(str));
		
		return(NULL);
	}

	for(i = 0; inb->par[i] >= '0' && inb->par[i] <= '9'; i++);

	if(inb->par[i])  /* alphanumeric input */ 
	{
		do
		{
			if(tbuf->jobid != NULL && strcmp(tbuf->jobid, inb->par) == 0)
				break;

			tbuf = tbuf->nblock;

			if(tbuf != NULL && tbuf->stblock)
				tbuf = NULL;
				
		} while(tbuf != NULL);
	}
	else
	{
		job = atoi(inb->par);

		if(job)
		{
			for(i = 1; i < job; i++)
			{
				tbuf = tbuf->nblock;

				if(tbuf != NULL && tbuf->stblock)
					tbuf = NULL;

				if(tbuf == NULL)
					break;
			}
		}
		else 
			tbuf = NULL;
	}
	return(tbuf);
}	

void outpid(CMD *inb)
{
	char *username;
	CONFIG *tbuf;
	char *str;
	
	if((tbuf = findblk(inb, &username)) == NULL || ! tbuf->pid)
		return;
		
	str = cpstr(20, "%d", (int)tbuf->pid);
	write(svsock, str, strlen(str));
}

void runcmd(CMD *inb)   /* run command */
{
	char *username;
	CONFIG *tbuf;
	int n1,n2,i;
	char inchr;
	char *jname,*str=NULL;

	if(sscanf(inb->par, "%d-%d%c", &n1, &n2, &inchr) == 2 && n1 > 0 && n2 > 0 && n1 <= n2)  /* range */
		*strchr(inb->par, '-') = 0;
	else 
		n1 = 0;

	if((tbuf = findblk(inb, &username)) == NULL)
	{
		str = cpstr(256, "No such job (%s)\n", inb->par);
		write(svsock, str, strlen(str));
		
		return;
	}

	if(n1)
	{
		i = n1;
		do
		{
			storequeue(tbuf, 1);
			
			i++;
			tbuf = tbuf->nblock;

			if(tbuf == NULL || tbuf->stblock)
				break;
		}
		while(i <= n2);

		str = cpstr(256, "Started range %d-%d\n", n1, i-1);
		write(svsock, str, strlen(str));

		return;
	}

	jname = (tbuf->jobid)?(tbuf->jobid):(inb->par);

	if(tbuf->pid)
	{
		str = cpstr(256, "Job already running (%s)\n", jname);
		write(svsock, str , strlen(str));
		
		return;
	}

	flushjob(tbuf);

	str = cpstr(256, "Starting command %s\n", jname);
	write(svsock, str, strlen(str));

	if(tbuf->swpdelay)
		tbuf->nextexec = runtab(tbuf, time(NULL));
		
	run_procs(tbuf);
}

int create_socket(char *sockname)
{
	int sock;
	struct sockaddr_un name;

	unlink(sockname);

	if((sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
		return(-1);
	
	name.sun_family = AF_UNIX;
	strcpy(name.sun_path, sockname);

	if(bind(sock, (struct sockaddr *)&name, offsetof(struct sockaddr_un, sun_path)+strlen(name.sun_path)+1) < 0)
		return(-1);

	chmod(sockname, S_IRUSR|S_IWUSR);
	listen(sock,1);
	fcntl(sock, F_SETFD, 1);
	return(sock);
}

void read_fifo(void)    /* read a command from socket */
{
	ssize_t nrd;
	CMD inb;

	signal(SIGALRM, (void *)sig_nop);
	alarm(15);

	if((nrd = read(svsock, &inb, sizeof(CMD))) < 0)
	{
		log_warn((errno == EINTR)?("timeout on socket"):("socket error while reading: %m"));
		return;
	}

	alarm(0);
	
	if(nrd != sizeof(CMD))
	{
		log_warn("invalid block length for input command");
		return;
	}

	switch(inb.cmd)
	{
		case CMD_UPDATE:
					update_ctab(inb.ctab+2);
					break;
		case CMD_INFO:
					outfifo(&inb);
					break;
		case CMD_RUN:
					runcmd(&inb);
					break;
		case CMD_PID:
					outpid(&inb);
					break;
		default:
					log_warn("unknown command received");
	}
}
